import argparse
import json
import inspect
import logging

from polar_route import __version__ as version
from polar_route.utils import setup_logging, timed_call, convert_decimal_days
from polar_route.mesh_generation.mesh_builder import MeshBuilder
from polar_route.vessel_performance.vessel_performance_modeller import VesselPerformanceModeller
from polar_route.route_planner import RoutePlanner
from polar_route.mesh_generation.environment_mesh import EnvironmentMesh
from polar_route.route_calc import route_calc


@setup_logging
def get_args(
        default_output: str,
        config_arg: bool = True,
        mesh_arg: bool = False,
        waypoints_arg: bool = False,
        format_arg: bool = False,
        format_conf: bool = False):
    """
    Adds required command line arguments to all CLI entry points.

    Args:
        default_output (str): The default output file location.
        config_arg (bool): True if the CLI entry point requires a <config.json> file. Default is True.
        mesh_arg (bool): True if the CLI entry point requires a <mesh.json> file. Default is False.
        waypoints_arg (bool): True if the CLI entry point requires a <waypoints.csv> file. Default is False.

    Returns:

    """
    ap = argparse.ArgumentParser()

    ap.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=version))

    # Optional arguments used in all CLI entry points
    ap.add_argument("-o", "--output",
                    default=default_output,
                    help="Output file")
    ap.add_argument("-v", "--verbose",
                    default=False,
                    action="store_true",
                    help="Turn on DEBUG level logging")

    if config_arg:
        ap.add_argument("config", type=argparse.FileType("r"), 
                    help="File location of a <config.json> file")

    if mesh_arg:
        ap.add_argument("mesh", type=argparse.FileType("r"),
                    help="File location of the environmental mesh")

    if waypoints_arg:
        ap.add_argument("waypoints", type=argparse.FileType("r"))

        # Optional arguments used when route planning.
        ap.add_argument("-p", "--path_only",
                        default=False,
                        action = "store_true",
                        help="output only the calculated paths")

        ap.add_argument("-d", "--dijkstra",
                        default=False,
                        action = "store_true",
                        help="output only the calculated paths")
        
    if format_arg:
        ap.add_argument("format",
                        help = "Export format to transform a mesh into. Supported \
                        formats are JSON, GEOJSON, Tif")
        ap.add_argument( "-f", "--format_conf",
                        default = None,
                        help = "File location of Export to Tif configuration parameters")



    return ap.parse_args()

@timed_call
def rebuild_mesh_cli():
    """
        CLI entry point for rebuilding the mesh based on its encoded config files.
    """

    default_output = "rebuild_mesh.output.json"
    args = get_args(default_output, mesh_arg=True, config_arg=False)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    mesh_json = json.load(args.mesh)
    config = mesh_json['config']

    # rebuilding mesh...
    rebuilt_mesh = MeshBuilder(config).build_environmental_mesh()
    rebuilt_mesh_json = rebuilt_mesh.to_json()

    if 'vessel_info' in config.keys():
        vessel_config = config['vessel_info']
        vp = VesselPerformanceModeller(rebuilt_mesh_json, vessel_config)
        vp.model_accessibility()
        vp.model_performance()
        rebuilt_mesh_json = vp.to_json()

    logging.info("Saving mesh to {}".format(args.output))
    
    json.dump(rebuilt_mesh_json, open(args.output, "w"), indent=4)



@timed_call
def create_mesh_cli():
    """
        CLI entry point for the mesh construction
    """
    
    default_output = "create_mesh.output.json"
    args = get_args(default_output)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    config = json.load(args.config)

    # Discrete Meshing
    cg = MeshBuilder(config).build_environmental_mesh()

    logging.info("Saving mesh to {}".format(args.output))
    info = cg.to_json()
    json.dump(info, open(args.output, "w"), indent=4)


@timed_call
def add_vehicle_cli():
    """
        CLI entry point for the vessel performance modeller
    """

    default_output = "add_vehicle.output.json"
    args = get_args(default_output, config_arg=True, mesh_arg=True)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    mesh_json = json.load(args.mesh)
    vessel_config = json.load(args.config)

    vp = VesselPerformanceModeller(mesh_json, vessel_config)
    vp.model_accessibility()
    vp.model_performance()

    logging.info("Saving mesh to {}".format(args.output))
    info = vp.to_json()
    json.dump(info, open(args.output, "w"), indent=4)


@timed_call
def optimise_routes_cli():
    """
        CLI entry point for the route optimisation
    """

    args = get_args("optimise_routes.output.json",
                    config_arg=True, mesh_arg=True ,waypoints_arg= True)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    if args.path_only:
        logging.info("outputting only path to {}".format(args.output))
    else: 
        logging.info("outputting full mesh to {}".format(args.output))

    rp = RoutePlanner(args.mesh.name, args.config.name, args.waypoints.name)
    
    rp.compute_routes()
    info_dijkstra = rp.to_json()
    
    if args.dijkstra:
        if args.path_only:
            json.dump(info_dijkstra['paths'], open('{}_dijkstra.json'.format('.'.join(args.output.split('.')[:-1])), 'w'), indent=4)
        else:
            json.dump(info_dijkstra, open('{}_dijkstra.json'.format('.'.join(args.output.split('.')[:-1])), 'w'), indent=4)
    
    rp.compute_smoothed_routes()
    info = rp.to_json()

    if args.path_only:
        json.dump(info['paths'], open(args.output, 'w'), indent=4)
    else:
        json.dump(info, open(args.output, "w"), indent=4)

@timed_call
def export_mesh_cli():
    """
        CLI entry point for exporting a mesh to standard formats.
        Currently supported formats are JSON, GEOJSON, TIF
    """
    # Default, used only by the Mesh Builder and PolarRoute
    args = get_args("mesh.json", 
                    config_arg = False, 
                    mesh_arg = True, 
                    format_arg = True)
        
    if args.format.upper() == "GEOJSON":
        args = get_args("mesh_geo.json", 
                    config_arg = False, 
                    mesh_arg = True, 
                    format_arg = True)
        
    elif args.format.upper() == "TIF":
        args = get_args("mesh.tif", 
                    config_arg = False, 
                    mesh_arg = True, 
                    format_arg = True)
    
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    mesh = json.load(args.mesh)
    env_mesh = EnvironmentMesh.load_from_json(mesh)

    logging.info(f"exporting mesh to {args.output} in format {args.format}")

    env_mesh.save(args.output, args.format , args.format_conf)

@timed_call
def calculate_route_cli():
    """
        CLI entry point to calculate the cost of a manually defined route within an existing mesh.
    """
    args = get_args("calculated_route.json",
                    config_arg = False, mesh_arg = True, waypoints_arg = True)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    logging.info(f"Calculating the cost of route {args.waypoints.name} from mesh {args.mesh.name}")

    calc_route = route_calc(args.waypoints.name, args.mesh.name)

    if calc_route is not None:
        max_time = convert_decimal_days(calc_route["features"][0]["properties"]["traveltime"][-1])
        max_fuel = round(calc_route["features"][0]["properties"]["fuel"][-1],2)

        logging.info(f"Calculated route has travel time: {max_time} and fuel cost: {max_fuel} tons")

        logging.info(f"Saving calculated route to {args.output}")
        with open(args.output, "w") as f:
            json.dump(calc_route, f, indent=4)






from polar_route.dataloaders.scalar.abstract_scalar import ScalarDataLoader
from polar_route.utils import date_range

from datetime import datetime
import logging
import numpy as np
import xarray as xr

class ThicknessDataLoader(ScalarDataLoader):
    def import_data(self, bounds):
        '''
        Creates a simulated dataset of sea ice thickness based on 
        scientific literature.
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            thickness_xr (xarray): 
                Sea Ice Density dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'thickness'
        '''
        # Look up table parameters hardcoded
        thicknesses = {'Ross': {'w': 0.72, 'sp': 0.67, 'su': 1.32, 'a': 0.82, 'y': 1.07},
                    'Bellinghausen': {'w': 0.65, 'sp': 0.79, 'su': 2.14, 'a': 0.79, 'y': 0.90},
                    'Weddell E': {'w': 0.54, 'sp': 0.89, 'su': 0.87, 'a': 0.44, 'y': 0.73},
                    'Weddell W': {'w': 1.33, 'sp': 1.33, 'su': 1.20, 'a': 1.38, 'y': 1.33},
                    'Indian': {'w': 0.59, 'sp': 0.78, 'su': 1.05, 'a': 0.45, 'y': 0.68},
                    'West Pacific': {'w': 0.72, 'sp': 0.68, 'su': 1.17, 'a': 0.75, 'y': 0.79},
                    'None': {'w': 0.72, 'sp': 0.67, 'su': 1.32, 'a': 0.82, 'y': 1.07}}
        seasons = {1: 'su', 2: 'su', 3: 'a', 4: 'a', 5: 'a', 6: 'w', 7: 'w', 8: 'w', 9: 'sp', 10: 'sp', 11: 'sp',
                12: 'su'}

        def ice_thickness(d, long):
            """
                Returns ice thickness. Data taken from 
                Table 3 in: doi:10.1029/2007JC004254
                Values don't extend to the northern hemisphere, 
                but are used for the time being as a placeholder
            """
            # The table has missing data points for Bellinghausen Autumn and Weddell W Winter, may require further thought
            month = d.month
            season = seasons[month]
            sea = None

            if -130 <= long < -60:
                sea = 'Bellinghausen'
            elif -60 <= long < -45:
                sea = 'Weddell W'
            elif -45 <= long < 20:
                sea = 'Weddell E'
            elif 20 <= long < 90:
                sea = 'Indian'
            elif 90 <= long < 160:
                sea = 'West Pacific'
            elif (160 <= long < 180) or (-180 <= long < -130):
                sea = 'Ross'
            else:
                sea = 'None'

            return thicknesses[sea][season]

        logging.info("- Setting boundaries for simulated dataset")
        start_date = datetime.strptime(bounds.get_time_min(), "%Y-%m-%d").date()
        end_date = datetime.strptime(bounds.get_time_max(), "%Y-%m-%d").date()

        lats = [lat for lat in np.arange(bounds.get_lat_min(), bounds.get_lat_max(), 0.05)]
        lons = [lon for lon in np.arange(bounds.get_long_min(), bounds.get_long_max(), 0.05)]
        dates = [single_date for single_date in date_range(start_date, end_date)]

        logging.info("- Generating dataset from boundaries")
        thick_data = xr.DataArray(
            data=[[[ice_thickness(dt, lng)
                    for lng in lons]
                for _ in lats]
                for dt in dates],
            coords=dict(
                lat=lats,
                long=lons,
                time=[dt.strftime("%Y-%m-%d") for dt in dates],
            ),
            dims=("time", "lat", "long"),
            name="thickness",
        )

        thick_df = thick_data.\
            to_dataframe().\
            reset_index().\
            set_index(['lat', 'long', 'time'])
        
        thickness_xr = thick_df.to_xarray()

        logging.debug("returning {} datapoints".format(len(thick_df.index)))

        del thick_df
        # No need to trim data, as was defined by bounds

        return thickness_xr



from polar_route.dataloaders.scalar.amsr import AMSRDataLoader
from polar_route.dataloaders.scalar.baltic_sea_ice import BalticSeaIceDataLoader
from polar_route.dataloaders.scalar.bsose_depth import BSOSEDepthDataLoader
from polar_route.dataloaders.scalar.bsose_sea_ice import BSOSESeaIceDataLoader
from polar_route.dataloaders.scalar.baltic_sea_ice import BalticSeaIceDataLoader
from polar_route.dataloaders.scalar.gebco import GEBCODataLoader
from polar_route.dataloaders.scalar.icenet import IceNetDataLoader
from polar_route.dataloaders.scalar.modis import MODISDataLoader
from polar_route.dataloaders.scalar.scalar_csv import ScalarCSVDataLoader
from polar_route.dataloaders.scalar.scalar_grf import ScalarGRFDataLoader
from polar_route.dataloaders.scalar.shape import ShapeDataLoader

from polar_route.dataloaders.vector.baltic_current import BalticCurrentDataLoader
from polar_route.dataloaders.vector.era5_wind import ERA5WindDataLoader
from polar_route.dataloaders.vector.north_sea_current import NorthSeaCurrentDataLoader
from polar_route.dataloaders.vector.oras5_current import ORAS5CurrentDataLoader
from polar_route.dataloaders.vector.sose import SOSEDataLoader
from polar_route.dataloaders.vector.vector_csv import VectorCSVDataLoader
from polar_route.dataloaders.vector.vector_grf import VectorGRFDataLoader

from polar_route.dataloaders.scalar.density import DensityDataLoader
from polar_route.dataloaders.scalar.thickness import ThicknessDataLoader

from glob import glob
import os


class DataLoaderFactory:
    '''
    Produces initialised DataLoader objects that can be used by the mesh to 
    quickly retrieve values within a boundary.
    '''
    @staticmethod
    def get_dataloader(name, bounds, params, min_dp=5):
        '''
        Creates appropriate dataloader object based on name
        
        Args:
            name (str): 
                Name of data source/type. Must be one of following - 
                'scalar_csv', 'scalar_grf', 'binary_grf', 'amsr', 'bsose_sic',
                'bsose_depth', 'baltic_sic', 'gebco', 'icenet', 'modis', 
                'thickness', 'density', 'circle', 'square', 'gradient',
                'checkerboard', 'vector_csv', 'vector_grf', 'baltic_currents',
                'era5_wind', 'northsea_currents', 'oras5_currents', 'sose'
            bounds (Boundary): 
                Boundary object with initial mesh space&time limits
            params (dict): 
                Dictionary of parameters required by each dataloader
            min_dp (int):  
                Minimum datapoints required to get homogeneity condition

        Returns:
            (Scalar/Vector/LUT DataLoader): 
                DataLoader object of correct type, with required params set 
        '''
        # Cast name to lowercase to make case insensitive
        name = name.lower()
        # Translate 'file' or 'folder' into 'files' key
        params = DataLoaderFactory.translate_file_input(params)
        
        # Add loader name to params
        params['dataloader_name'] = name
        params['min_dp'] = min_dp
        
        dataloader_requirements = {
            # Scalar
            'scalar_csv':   (ScalarCSVDataLoader, ['files']),
            'scalar_grf':   (ScalarGRFDataLoader, ['binary']),
            'binary_grf':   (ScalarGRFDataLoader,['binary']),
            'amsr':         (AMSRDataLoader, ['files', 'hemisphere']),
            'bsose_sic':    (BSOSESeaIceDataLoader, ['files']),
            'bsose_depth':  (BSOSEDepthDataLoader, ['files']),
            'baltic_sic':   (BalticSeaIceDataLoader, ['files']),
            'gebco':        (GEBCODataLoader, ['files']),
            'icenet':       (IceNetDataLoader, ['files']),
            'modis':        (MODISDataLoader, ['files']),
            'thickness':    (ThicknessDataLoader, []),
            'density':      (DensityDataLoader, []),
            # Scalar - Abstract shapes
            'circle':       (ShapeDataLoader, []),
            'square':       (ShapeDataLoader, []),
            'gradient':     (ShapeDataLoader, []),
            'checkerboard': (ShapeDataLoader, []),
            # Vector
            'vector_csv':       (VectorCSVDataLoader, ['files']),
            'vector_grf':       (VectorGRFDataLoader, []),
            'baltic_currents':  (BalticCurrentDataLoader, ['files']),
            'era5_wind':        (ERA5WindDataLoader, ['files']),
            'northsea_currents':(NorthSeaCurrentDataLoader, ['files']),
            # TODO make it run from 'files'
            'oras5_currents':   (ORAS5CurrentDataLoader, ['files']),
            'sose':             (SOSEDataLoader, ['files'])

        }
        # If name is recognised as a dataloader
        if name in dataloader_requirements:
            # Set data loader and params required for it to work
            data_loader = dataloader_requirements[name][0]
            required_params = dataloader_requirements[name][1]
        else: 
            raise ValueError(f'{name} not in known list of DataLoaders')

        # Assert dataloader will get all required params to work
        assert all(key in params for key in required_params), \
            f'Dataloader {name} is missing some parameters! Requires {required_params}. Has {list(params.keys())}'

        # Create instance of dataloader
        return data_loader(bounds, params)
    
    @staticmethod
    def translate_file_input(params):
        '''
        Allows flexible file specification in params. Translates 'file' or 
        'folder' into 'files'
        
        Args:
            params (dict): 
                Dictionary of parameters written in config
        '''
        if 'file' in params:
            params['files'] = [params['file']]
            del params['file']
        elif 'folder' in params:
            folder = os.path.join(params['folder'], '') # Adds trailing slash if non-existant
            params['files'] = sorted(glob(folder+'*'))
            del params['folder']
        return params


region_schema = {
    "type": "object",
    "required": ["latMin",  "latMax", 
                 "longMin", "longMax",
                 "startTime", "endTime",
                 "cellWidth", "cellHeight"],
    "additionalProperties": False,
    "properties":{    
        "latMin": {"type": "number"},
        "latMax": {"type": "number"},
        "longMin": {"type": "number"},
        "longMax": {"type": "number"},
        "startTime": {"type": "string"},
        "endTime": {"type": "string"},
        "cellWidth": {"type": "number"},
        "cellHeight": {"type": "number"}
    }
}

dataloader_schema = {
    "type": "object",
    "required": ["loader", "params"],
    "additionalProperties": False,
    "properties":{
        "loader": {"type": "string"},
        "params": {"type": "object"}
        
    }
}

splitting_schema = {
    "type": "object",
    "required": ["split_depth", "minimum_datapoints"],
    "additionalProperties": False,
    "properties":{
        "split_depth": {"type": "integer"},
        "minimum_datapoints": {"type": "integer"}
    } 
}

mesh_schema = {
    "type": "object",
    "required": ["Mesh_info"],
    "additionalProperties": False,
    "properties":{
        "Mesh_info": {
            "type": "object",
            "required": ["Region", "Data_sources", "splitting"],
            "additionalProperties": False,
            "properties":{
                "Region": {
                    "$ref": "#/region_schema"
                },
                "Data_sources": {
                    "type": "array",
                    "items": {
                        "$ref": "#/dataloader_schema"
                    },
                },
                "splitting": {
                    "$ref": "#/splitting_schema"
                }
            }
        }
    },
    "region_schema": region_schema,
    "dataloader_schema": dataloader_schema,
    "splitting_schema": splitting_schema
}


from polar_route.dataloaders.scalar.abstract_scalar import ScalarDataLoader
from polar_route.utils import date_range

from datetime import datetime
import logging

import numpy as np
import xarray as xr

class DensityDataLoader(ScalarDataLoader):
    def import_data(self, bounds):
        '''
        Creates a simulated dataset of sea ice density based on 
        scientific literature.
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            density_xr (xarray): 
                Sea Ice Density dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'density'
        '''
        # Look up table parameters hardcoded
        seasons = {
            1: 'su', 2: 'su', 12: 'su',
            3: 'a', 4: 'a', 5: 'a', 
            6: 'w', 7: 'w', 8: 'w', 
            9: 'sp', 10: 'sp', 11: 'sp',
            }
        densities = {
            'su': 875.0, 
            'sp': 900.0, 
            'a': 900.0, 
            'w': 920.0
            }

        def ice_density(d):
            '''
            Retrieve ice density from a datetime object
            '''
            month = d.month
            season = seasons[month]
            den = densities[season]
            return den

        # Set boundary coordinates and dates
        logging.info("- Setting boundaries for simulated dataset")
        start_date = datetime.strptime(bounds.get_time_min(), "%Y-%m-%d").date()
        end_date = datetime.strptime(bounds.get_time_max(), "%Y-%m-%d").date()

        lats = [lat for lat in np.arange(bounds.get_lat_min(), bounds.get_lat_max(), 0.05)]
        lons = [lon for lon in np.arange(bounds.get_long_min(), bounds.get_long_max(), 0.05)]
        dates = [single_date for single_date in date_range(start_date, end_date)]

        # Generate a density dataset from hardcoded values
        logging.info("- Generating dataset from boundaries")
        density_data = xr.DataArray(
            data=[[[ice_density(dt)
                    for _ in lons]
                for _ in lats]
                for dt in dates],
            coords=dict(
                lat=lats,
                long=lons,
                time=[dt.strftime("%Y-%m-%d") for dt in dates],
            ),
            dims=("time", "lat", "long"),
            name="density",
        )
        # Cast as dataframe
        density_df = density_data.\
            to_dataframe().\
            reset_index().\
            set_index(['lat', 'long', 'time'])
        # Cast to xr.Dataset for faster processing
        density_xr = density_df.to_xarray()
        # Clear memory of now unused dataframe
        del density_df
        # No need to trim data, as was defined by bounds

        return density_xr


from polar_route.dataloaders.vector.abstract_vector import VectorDataLoader

import logging

import xarray as xr

from datetime import datetime

class ERA5WindDataLoader(VectorDataLoader):
    def import_data(self, bounds):
        '''
        Reads in data from a ERA5 NetCDF file. 
        Renames coordinates to 'lat' and 'long'
        
        Args:
            bounds (Boundary): Initial boundary to limit the dataset to
            
        Returns:
            xr.Dataset: 
                ERA5 wind dataset within limits of bounds. 
                Dataset has coordinates 'lat', 'long', and variable 'elevation'
        '''
        # Open Dataset
        if len(self.files) == 1:    data = xr.open_dataset(self.files[0])
        else:                       data = xr.open_mfdataset(self.files)
        # Change column names
        data = data.rename({'latitude': 'lat',
                            'longitude': 'long'})

        # Set min time to start of month to ensure we include data as only have
        # monthly cadence. Assuming time is in str format
        time_min = datetime.strptime(bounds.get_time_min(), '%Y-%m-%d')
        time_min = datetime.strftime(time_min, '%Y-%m-01')

        # Reverse order of lat as array goes from max to min
        data = data.reindex(lat=data.lat[::-1])

        # Trim to initial datapoints
        data = self.trim_datapoints(bounds, data=data)
        
        return data



from polar_route.vessel_performance.abstract_vessel import AbstractVessel
from polar_route.mesh_generation.aggregated_cellbox import AggregatedCellBox


class UnderwaterVessel(AbstractVessel):
    """
        Class to model the performance of an underwater vessel
    """

    def __init__(self, params):
        """
            Args:
                params (dict): vessel parameters from the vessel config file
        """
        self.vessel_params = params

    def model_performance(self, cellbox):
        """
            Method to determine the performance characteristics for the underwater vessel
                Args:
                    cellbox (AggregatedCellBox): input cell from environmental mesh
        """
        pass

    def model_accessibility(self, cellbox):
        """
            Method to determine if a given cell is accessible to the underwater vessel
                 Args:
                    cellbox (AggregatedCellBox): input cell from environmental mesh
        """
        pass

